import * as React from "react";
import { useRef, createElement, useContext, useState, useEffect } from "react";
import { EditorView } from "@codemirror/view";
function createScheduler() {
  let task = [];
  return {
    request: (callback) => {
      if (!task.length) {
        queueMicrotask(() => task.shift()());
      }
      task = [callback];
    },
    idle: () => !task.length
  };
}
function createSubject(_value) {
  const observers = /* @__PURE__ */ new Set();
  return {
    get value() {
      return _value;
    },
    getValue: () => _value,
    next: (value) => {
      if (_value !== value) {
        _value = value;
        [...observers].forEach((next) => next(value));
      }
    },
    subscribe: (next) => {
      observers.add(next);
      next(_value);
      return () => observers.delete(next);
    }
  };
}
function createCodeMirror(initialConfig) {
  const config$ = createSubject(initialConfig);
  function getViewConfig(prevState) {
    const config = config$.value;
    return typeof config === "function" ? config(prevState) : { ...config };
  }
  const view$ = createSubject(null);
  const container$ = createSubject(null);
  function createView(container) {
    var _a;
    const prevState = (_a = view$.value) == null ? void 0 : _a.state;
    return new EditorView({
      ...getViewConfig(prevState),
      parent: container
    });
  }
  const scheduler = createScheduler();
  container$.subscribe((container) => {
    scheduler.request(() => {
      var _a;
      (_a = view$.value) == null ? void 0 : _a.destroy();
      view$.next(container && createView(container));
    });
  });
  config$.subscribe(() => {
    if (scheduler.idle()) {
      scheduler.request(() => {
        var _a;
        const container = container$.value;
        if (container) {
          (_a = view$.value) == null ? void 0 : _a.destroy();
          view$.next(createView(container));
        }
      });
    }
  });
  return {
    getView: view$.getValue,
    subscribe: view$.subscribe,
    setContainer: container$.next,
    setConfig: config$.next
  };
}
function createContext() {
  const CodeMirrorContext = React.createContext(null);
  const CodeMirrorProvider = ({ config, children }) => {
    const instanceRef = useRef(null);
    function getInstance() {
      if (!instanceRef.current) {
        instanceRef.current = createCodeMirror(config);
      }
      return instanceRef.current;
    }
    return createElement(CodeMirrorContext.Provider, { value: getInstance() }, children);
  };
  function useCodeMirrorContext() {
    const instance = useContext(CodeMirrorContext);
    if (!instance) {
      throw new Error(
        "could not find instance from CodeMirrorContext; please ensure the component is wrapped in a <Provider>"
      );
    }
    return instance;
  }
  return {
    Provider: CodeMirrorProvider,
    useCodeMirror: useCodeMirrorContext
  };
}
function useView({ getView, subscribe }) {
  const [view, setView] = useState(getView);
  useEffect(() => subscribe(setView), [subscribe]);
  return view;
}
function defineViewEffect(setup) {
  return setup;
}
function useViewEffect({ subscribe }, setup) {
  useEffect(() => {
    function setupEffect(view) {
      return view && setup(view);
    }
    let cleanup;
    const unsubscribe = subscribe((view) => {
      cleanup == null ? void 0 : cleanup();
      cleanup = setupEffect(view);
    });
    return () => {
      unsubscribe();
      cleanup == null ? void 0 : cleanup();
    };
  }, [setup, subscribe]);
}
function create(config) {
  const cm = createCodeMirror(config);
  return {
    ...cm,
    useView: () => useView(cm),
    useViewEffect: (setup) => useViewEffect(cm, setup)
  };
}
export {
  create,
  createCodeMirror,
  createContext,
  defineViewEffect,
  useView,
  useViewEffect
};
//# sourceMappingURL=index.js.map
